"""
    This file contains constants that are generally used by the program
"""

#physical constants

kb = 1.38064852e-23  #Boltzmann constant J/K
me = 9.10938356e-31  #mass of electron kg
h = 6.62607004e-34  #Planck's constant J*S
Nav = 6.02214086e23  #Avogadro's number
Atometer = 1.0e-10  #conversion of angstroms to meters
